# activestate perl for windows
#!/usr/bin/perl

###############################################################
## Author:  Dan Barrett a.k.a. wily duck (wilyduck @digitalamusement com)
## Date:    6/8/2011 (original)
## Program: versioneer.pl
## Purpose: Automatically progress a map version.
##
## Important: This script is a README, manual, and script all-in-one. Read it through so you know
##            exactly what's going on! It's really not that hard, but I make sure you're sure.
##
## Requirements: You need perl to execute this script. For windows, get activestate perl at:
##               http://www.activestate.com/activeperl/downloads
##               Once installed, you don't need to do anything else to configure perl. Automagic!
##
## Notes:   1. Read the "Info" and "Instructions" section. This script is pretty well documented.
##          2. It does NOT search for shader dependencies, nor does it reference external shaders -
##             meaning that this script works great for stand-alone maps. If your map does reference
##             external resources, you can optionally specify them in Section 2 per instructions, or
##             leave them be if you want them to continue referencing their current resource.
##          3. The versioneer only copies files - except for copysame() which rewrites a file.
##          4. No undo, so worse case, you'll end up with a new map you didn't want. Won't overwrite
##             an existing .map - in fact we'll abort if you accidentally specified an existing .map.
##             that already exists. You should be backing up your .map's anyway! Daily!
##          5. Can be re-run immediately if say a folder didn't copy correctly and you fixed
##             that folder. Delete the new .map created and re-run the versioneer.
##          6. Built for perl on windows. Easily portable to *nix if you DIY. :)
##          7. I trust this script for my maps, so you should have some faith the versioneer won't
##             break your maps either. Even if you Do It Wrong, you'll end up with extra files, maybe
##             a shaderlist.txt with an extra line. Not horrible.
##          8. When the versioneer packages a map, it does not create a final pk3 file. It bundles all files
##             that it copies into a package directory for you to verify, Zip, and rename to .pk3 yourself.
##          9. You will need to recompile your .bsp to match resources in your new map.
##
## Version: v1.5 - 04072013.2 - Added automatic model migration for making your map more stand-alone.
##          v1.4 - 04072013.1 - Added packaging logic. No longer need to specify $path with all functions.
##                              Made copying more safe (against overwrites if recopying to a new map).
##          v1.3 - 04062013.0 - Added a bunch more comments for other users that might find the
##                              versioneer useful! Added more logic to copysame(). Now uses
##                              $path so the versioneer can live anywhere on your system you like.
##                              Added all sorts of user-friendliness and yet more logic!
##          v1.2 - 06142011.0 - Added safety check to verify if new map already exists. Copying
##                              the .map file is now the last action.
##          v1.1 - 06112011.0 - Restructured the configuration sections. Lots more commenting.
##                              Added second arg to &copydir. No longer need to double backslash
##                              folder names.
##          v1.0 - 06082011.1 - Works just fine. Should have made this years ago.
##          v0.0 - 06082011.0 - Starting new script from newscript template 03262010.2
##
## Bugs:    #0 - 06082011.0 - None.
##
## To-do:   1. None.
##
## Warranty: None! You're on your own!
##
## Support: Send me an email if you have questions. I'll usually get back to you within 48 hours. Not
##          guaranteed of course! /legaldisclaimer
##
## License: Open source, freely redistributable, modifications and forks are expected. Give me credit
##          as the original author including name and email.
##
## Usage:   perl versioneer.pl
###############################################################



### Info! ###

# Really handy script to fork a copy of your map as a new stand-alone map. Great for beta's, new point releases,
# or trying something new. For example, ut4_yourmap is released, now you want to move on to ut4_yourmap2.
# The versioneer will copy the resources of ut4_yourmap to ut4_yourmap2, so you won't mess with the
# original map. Even better, say you have ut4_yourmap_beta, and you want to create a new map that doesn't
# conflict with old shaders/textures/etc. Fork a ut4_yourmap_beta2! All new, _completely independent_ of
# the first map. Don't forget to remove that old pk3 from q3ut4 eventually though. Remember UrT/Q3 has
# to parse every pk3 each time you load UrT!
#
# The versioneer really appreciates when you treat your map as a stand-alone map, free of referencing
# shaders/textures/sounds/models/etc from UrT pack files or other maps. This way, you have complete control
# over your maps' resources. You also don't inadvertantly squash someone else's textures/shaders. It's
# good practice. It's also fun to play your map in a completely different mod, like excessive or padman!
# If you don't know how to build an independent map, it's simple. Don't use anyone else's resources.
# Don't call any other shaders than say "textures/ut4_yourmap/shadername" in your ut4_yourmap.shader file.
# All models you use reside in models/mapobjects/ut4_yourmap/modelname". Textures reside in
# textures/ut4_yourmap/texture.tga  Etc!


### Instructions ###

# The versioneer lives in your q3ut4/ directory by default, but as long as you specify $path, anywhere is fair game.
#
# 1. Modify variables in Section 1. Specify the source map being copied, and new map being created.
#
# 2. If the source map references external resources, i.e. aux shaders that do not match the source map name,
#    or any one-offs that should be addressed, you can have the script pull in these one-offs and merge
#    them into the new map.
#
# 3. Execute cmd: perl versioneer.pl
#
# 4. Validate your new map!


# Your original map resources will not be harmed! We copy from, never write to. (except for the copysame() function,
# normally used for shaderlist.txt





#########################
######  Section 1  ######
#########################




###############################
### Map names and path here ###
###############################

## Use / forward slashes in all cases, not \ backslashes!

# Where is your Urban Terror exe?
$path = "e:/games/fps/ioUrT4_dev";  # suffix slash / not necessary. we'll fix it if you demand to append it
$urtexe = "ioUrbanTerror.exe";  # name of the UrT exe file to make sure $path is specified correctly

# Copy from map1 and create map2.
$map1 = "ut4_superman2_beta";
$map2 = "ut4_superman2_beta2";

# Do you want your new map copied normally, or package it as a subdir under q3ut4/ ?
$package = "";  # Specify "yes" if you want to create $map2 resources in q3ut4/$packdir/$map2/ for package bundling.
                # Won't update shaderlist.txt if set to "yes". Set to "force" if you need it to.

$dotmap = "";   # Triggering $package will copy your .map to the packaging directory, so if you don't want to expose your
                    # map source file, specify "move" here to move it back to q3ut4/maps/, or specify "none" to skip copying
                    # map1 to a new map2 entirely. Leave blank for standard copying. Warning: If you specify "none", $standalone
                    # will not copy models or modify model paths in the .map!

$packdir = "dev/";  # If $package is "yes", this will tell us where to place the package files.
                    # Output path: $path/q3ut4/$packdir/$map2/

$standalone = "yes";  # "yes" for converting all references of models/mapobjects/ to models/mapobjects/$map2/ within the .map file.
                      # Super useful to copy out shared models into an isolated environment where you can shader them
                      # without inadvertantly affecting other maps. This also won't trigger if $map2 already exists.
                      # Funny story. Check out ut4_superman.pk3/scripts/vehicle_enhancements.shader to see how NOT to do it.
                      # Botched that one and HEARD about it when the shared van was overbright in another map. Ouch. Sure this
                      # fattens up your pk3, but in this age, an extra 2-10MB is nothing. You should splurge. Trim fat by not
                      # including unused resources.



#### End of Section 1 ####

## We'll run the standard functions, then get to additional functions. Move ahead one space to Section 2.
&startup();  # Prepare our environment.
&primary();  # Do the normal copy stuff first.






#########################
######  Section 2  ######
#########################



##############################################
### Non-default actions specified by user ####
##############################################

## Specify additional actions here. Definitions below. ##

&copydir("models/mapobjects/","recursive");  # Not the whole mapobjects dir lol! this path gets appended with $map2 during copy.
&copydir("env/","");
&copydir("levelshots/","");  # Animated levelshot! Has its own dir for levelshot frames.
&copydir("textures/","/editor");  # Radiant editor stuff shaders reference.

&createfile("scripts/","-ss_b1-models.shader");  # This will rename all $map1 instances, leaving "ut4_ss_b1" shader names alone.
&copysame("scripts/shaderlist.txt","add","-ss_b1-models") if $package ne "yes";  # Don't forget to update shaderlist! If not bundling.


&copyfile("maps/",".bsp") if $package eq "yes";  # remember to recompile after versioneer'ing to a new map.



## More examples of additional actions when I migrated ut4_temple2_alpha to ut4_temple2_beta

# &createfile("scripts/","-plant.shader");  # needed to progress ut4_temple2_alpha-plant.shader to ut4_temple2_beta-plant.shader
# &copysame("scripts/shaderlist.txt","-plant.shader");  # ...and needed to update shaderlist.txt accordingly
# &copydir("textures/","/tf1");  # had additional textures in textures/ut4_temple2_alpha/tf1/ that needed to migrate




########################
## Available actions: ##
########################

## Functions:

# createfile ([dir],[file_suffix]) - reads the mapname text within a file and copies the new mapname text to a new mapname
#                                    file [primarily for .map, .shader]
#                                    Note, this is a global search and replace, and if $map1 happens to be inadvertantly
#                                    a partial map name for a name that exists within the file (rare), you may have a problem!
#                                    i.e. $map1 = "ut4_yourmap" but your file contains "ut4_yourmaprocks" ... well you'll
#                                    find all sorts of $map2 in there like "ut4_newmaprocks". Make sure $map1 and $map2
#                                    are EXACTLY what you want to find and replace if you think there is a conflict!

# copyfile ([dir],[file_suffix]) - copies old mapname file to new mapname file [primarily for minimap, levelshot]

# copydir ([dir],[leaf]) - copies old mapname folder in a dir to new mapname in the same folder, and leaf folders
#                          under that dir if specified [primarily for textures/, sound/]
#                          Be sure to specify the prefix / in [leaf], i.e. "/foldername".
#                          Specify "recursive" for [leaf] if you want all subfolders copied as-is. This will overwrite
#                          destination files.

# copysame ([file],[action],[suffix]) - subs mapname text within a file and rewrites that file [primarily for shaderlist.txt]
#                              copysame() will ONLY copy within q3ut4/, not a packaging path. Thus best for shaderlist.txt.
#                              If [action] is "add" or null (default), copysame will simply insert $map2 into [file],
#                              preferably after $map1.
#                              If [action] is "strict", copysame will only replace instances where the line begins and
#                              ends with $map1 as the text with $map2.
#                              If [action] is "global", copysame will globally search and replace all instances of $map1
#                              with $map2 within [file].
#                              [suffix] allows for appending shader names, like mapname + suffix, i.e. ut4_yourmap-surfaces

## Arguments:
# ..where [file_suffix] is any text after the mapname, i.e. ut4_yourmap[.txt] to take action on ut4_yourmap.txt
# ..where [dir] is a dir that can either be blank to specify the default maps/ dir, or specified like ../scripts/
# ..where [file] is the dir and file name, i.e. $path/scripts/shaderlist.txt
# ..where [leaf] is a folder path suffix of a mapname dir, i.e. for textures/yourmap/editor you specify ("../textures/","/editor")






##############################################
#### End of standard script configuration ####
##############################################

&finish();

























# That means go away or prepare to wield your sword of +3 Perling!

################################################################################################################################
################################################################################################################################
################################################################################################################################

# From here down is code that you shouldn't mess with, unless you have a reason to. Reasonably
# commented where code behavior isn't obvious.









################################################
#### SUBS - no modifications required below ####
################################################

sub startup {

  die "Your path doesn't seem right. Can't find $urtexe" if ! -e "$path/$urtexe";

  $path .= "/" if $path !~ /\/$/;  # Add a / suffix if none specified
  $path .= "q3ut4/";

  die "Check your map name and try again. $	path/maps/$map1.map\n" if ! -e "$path/maps/$map1.map";  # Sanity check
  die "New .map file already exists! Aborting as a precaution. $destpath/maps/$map2.map\n" if -e "$destpath/maps/$map2.map";  # Safety check

  if ($package eq "yes") {
    $destpath = $path . $packdir . "$map2/";

  } else {  # ensure copy-from path then matches copy-to path if we're not packaging
    $destpath = $path;
  }

  print "\nThe versioneer will copy $map1 to $map2";
  print " as a package bundle in $destpath" if $package eq "yes";

  print "\n\nReady! Press enter to begin! (or control-C to abort)\n\n";
  <STDIN>;

## Unfortunately the next runcommand() executes no matter what, even when control-C is pressed, so this gets around that.
  &runcommand("echo", "The Versioneer");

  return 0;

}




### Primary copy actions ###

sub primary {

  if ($package eq "yes") {
    &op("Creating package bundle directory $destpath");

    if (-d "$destpath") {
      print " ** SKIPPING ** copying because the directory exists:\n$destpath\n";
    } else {
      &runcommand("mkdir", "$destpath");
    }
  }

  &createfile("scripts/",".shader");

  &createfile("scripts/",".arena");

  if ($package ne "yes") {
    &copysame("scripts/shaderlist.txt","add");  # Inject $map2 into shaderlist.txt
  }

  &copyfile("levelshots/",".tga");  # Looks for ut4_yourmap.tga, turns into ut4_newmap.tga

  &copydir("sound/","");

  &copydir("textures/","");

  &copyfile("maps/",".tga");

## Handle the .map action. Any mapper knows this is touchy business.
  if ($dotmap ne "none") {
     &op("Creating $map2.map");

    if ($dotmap eq "move") {  # It's necessary to move the map after creation because by the function, it will be created
                              # in $destpath, so we must move it back to $path specifically.

      if (! -e "$path/maps/$map2.map") {  # Check the .map doesn't already exist in the source map/ directory.
        &createfile("maps/",".map");

        &op("Moving $destpath/maps/$map2.map back to $path");
        &runcommand("move", "$destpath/maps/$map2.map", "$path/maps/$map2.map");

      } else {
        print " ** SKIPPING ** creating the .map, because the destination .map already exists:\n$path" . "maps/$map2.map\n";
      }

    } else {
      &createfile("maps/",".map");
    }

  }

## Note: Specify env/ or models/ changes in Section 2 above

  print "\nInfo: Primary functions complete. Additional functions next.\n\n";

  return 0;

}



sub finish {

   my $line, $modelpath = "";

## If we want to migrate the .map to a standalone version that does not reference global models
  if ($standalone) {
    print "\nModels found that are being migrated to stand-alone:\n";

    foreach $line (@newmodel) {

      $line =~ s/^\"model\" \"//;  # scrub the line
      $line =~ s/\"\s*$//;

      $models{$line}++;  # remove duplicates with a hash

    }

    foreach $modelpath (sort keys %models) {  # show what models we modified in the .map
      $modelpath =~ s/^(\S+)\/.*?$/$1/;  # strip the model file name from the path

      $destmodelpath = $modelpath;
      $destmodelpath =~ s/mapobjects\//mapobjects\/$map2\//;  # insert map2 into the model path

      &runcommand("xcopy /E /Y","$path$modelpath/*","$destpath$destmodelpath/*");

    }

    print "No external shared models to import.\n" if ! $modelpath;

  }

  print "\nComplete. The versioneer aims to please. Press any key to continue.\n"; `pause`;

  exit;

}



######################
## Function actions ##
######################

sub copydir {

  my ($dir,$leaf) = ();
  my ($dir,$leaf) = @_;

  $dir .= "/" if $dir !~ /\/$/;  # add a / suffix if not specified
  $leaf =~ s/\/$//;  # need to lose the suffixed / to copy properly

  if ($leaf eq "recursive") {  # xcopy has /Y set, so it will overwrite destination files

    &op("copydir recursive $dir$map2");

    &runcommand("mkdir", "$destpath$dir$map2");
    &runcommand("xcopy /E /Y", "$path$dir$map1", "$destpath$dir$map2");

  } else {

    &op("copydir $dir$map2$leaf");

    if (-d "$destpath$dir$map2$leaf") {
      print " ** SKIPPING ** copying because the directory exists:\n$destpath$dir$map2$leaf\n";
      return 0;
    }

    &runcommand("mkdir", "$destpath$dir$map2$leaf");
    &runcommand("copy", "$path$dir$map1$leaf/*", "$destpath$dir$map2$leaf/*");

  }

}


sub copyfile {

  my ($dir,$suf) = ();
  my ($dir,$suf) = @_;

  $dir .= "/" if $dir !~ /\/$/;  # add a / suffix if not specified

  &op("copyfile $dir$map2$suf");

  if (! -e "$path$dir$map1$suf") {
    print " ** SKIPPING ** copying because the source file doesn't exist:\n$path$dir$map1$suf\n";
    return 0;
  }

  if (-e "$destpath$dir$map2$suf") {
    print " ** SKIPPING ** copying because the destination file exists:\n$destpath$dir$map2$suf\n";
    return 0;
  }

  &runcommand("mkdir", "$destpath$dir") if ! -d "$destpath$dir";
  &runcommand("copy", "$path$dir$map1$suf", "$destpath$dir$map2$suf");

}


sub createfile {

  my ($dir,$suf) = ();
  ($dir,$suf) = @_;

  $dir .= "/" if $dir !~ /\/$/;  # add a / suffix if not specified

  &op("createfile $dir$map2$suf");

  if (! -e "$path$dir$map1$suf") {
    print " ** SKIPPING ** creating file because the source doesn't exist:\n$path$dir$map1$suf\n";
    return 0;
  }

  &op("Directory $destpath$dir needs to be created") if ! -d "$destpath$dir";
  &runcommand("mkdir", "$destpath$dir") if ! -d "$destpath$dir";

  if (-e "$destpath$dir$map2$suf") {
    print " ** SKIPPING ** copying because the destination file exists:\n$destpath$dir$map2$suf\n";
    return 0;
  }

  open (NEW, ">$destpath$dir$map2$suf") || die "Could not write $destpath$dir$map2$suf\n$!";
  open (MAP, "$path$dir$map1$suf") || die "Could not read $path$dir$map1$suf\n$!";
  while ($line = <MAP>) {
    $line =~ s/$map1/$map2/g;  # globally replaces instances of $map1 with $map2

## Here we can make all models stand-alone
    if ($standalone) {
      if ($line =~ /^\"model\"/) {

        if ($line !~ /\/$map2\//) {  # line already modified to map2, don't duplicate the location
          push (@newmodel, $line);  # track what models have been modified

          $line =~ s/mapobjects\//mapobjects\/$map2\//;  # insert map2 into the model path
        }
      }
    }

    print NEW $line;

  }

  close MAP;
  close NEW;

}


sub copysame {

  my ($file,$action,$suffix,$line) = ();
  ($file,$action,$suffix) = @_;

  $action = "add" if ! $action;  # set default action to add

  my $find = $map1 . $suffix;
  my $replace = $map2 . $suffix;

  &op("copysame $replace $action");

  if (! -e "$path$dir$file") {
    print " ** SKIPPING ** copying because the source doesn't exist:\n$path$dir$file\n";
    return 0;
  }

  my @contents = ();

  open (FILE, "$path$file") || (warn ("Couldn't read $path$file\n$!") && return 1);  # We will ONLY copy in and out of q3ut4/, not a packaging path
  while ($line = <FILE>) {
    $line =~ s/\s+$//;  # strip whitespace from end of line

    if ($action eq "global") {  # global string find and replace
      $line =~ s/$find/$replace/g;      

    } elsif ($action eq "strict") {  # only replaces lines that begin and end with the string
      if ($line eq $find) {
        $line = $replace;
      }

    } else {  # action is [add] or null
      if ($line eq $find) {
        $line .= "\n$replace";
        $action = "added";
      }

      if ($line eq $replace) {  # shouldn't see map2 in shaderlist.txt already - so don't dupe it
        print " ** SKIPPING ** Duplicate $replace is already in $file\n";
        @contents = ();

        return 0;
      }

    }

    push (@contents, $line);

  }
  close FILE;

  if ($action eq "add") {  # didn't find map1? Then append map2 to end of file
    push (@contents, $replace);
  }

  &runcommand("copy", "$path$file", "$path$file.old");  # keep the old file just in case!

## Write out the file
  open (FILE, ">$path$file") || (warn ("Couldn't write $path$file\n$!") && return 1);
  foreach $line (@contents) {
    next if $line !~ /\S+/;  # skip blank lines
    print FILE "$line\n";
  }
  close FILE;

  @contents = ();

}


## Reformats commands to work in the DOS cmd environment, then runs the command.
sub runcommand {

  my ($cmd,$arg1,$arg2) = ();
  ($cmd,$arg1,$arg2) = @_;

  $arg1 =~ s/\/$// if $cmd eq "mkdir";  # irritating DOS quirk, mkdir doesn't like trailing /

  $arg1 =~ s/\//\\/g;  # Replace /'s with \'s (and double backslash the \'s)
  $arg2 =~ s/\//\\/g if $arg2;

  $arg1 = "\"$arg1\"";
  $arg2 = "\"$arg2\"" if $arg2;

  $cmd .= " $arg1";
  $cmd .= " $arg2" if $arg2;

  print "cmd> $cmd\n";

  `$cmd`;  # run the command via cmd shell

  return 0;

}


## For making output more readable/sane
sub op {

  my $line = $_[0];

  print "\nop> $line ...\n";

  return 0;

}

